/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "ciao.h"
#include "iao_http.h"
#include "iao_threads.h"
#include "iao_base64.h"
#include "iao_config.h"

#ifdef WIN32
#pragma warning(disable : 4996)
#endif 

#define IAOP_TAG_OPEN "<iaop>"
#define IAOP_TAG_CLOSE "</iaop>"

#define IAOP_VERSION 1
#define IAOP_SERVLET "/iaop1"

static time_t iao_born_time;
static int iao_iteration = 0;
static char host[IAOP_MAX_LEN];
static int port;
static char proxyhost[IAOP_MAX_LEN];
static int proxyport;
static char proxycridentials[IAOP_MAX_LEN];
static int doproxy = 0;
static iao_mutex mx;

static int iao_firsttime()
{
	static int first_api_call=1;
	if (first_api_call==1)
	{
		char buff[256];
		iao_born_time = time(NULL);
		if (iao_tcpinit()!=0)
		{
			return -1;
		}
		mx = iao_createmutex();
		if (mx==NULL)
		{
			iao_logmsg("Can't create main mutex");
			return -1;
		}
		iao_configinit();

		if (iao_getembeddedvariable("RUON_HOST", host)==-1)
		{
#ifdef CIAO_TEST
			strcpy(host, "ruondev");
#else
			strcpy(host, "agent.r-u-on.com");
#endif
		}

		if (iao_getembeddedvariable("RUON_PORT", buff)==0)
		{
			port = atoi(buff);
		}
		else
		{
#ifdef IAO_SSL
#ifdef CIAO_TEST
			port = 8443;
#else
			port = 443;
#endif
#else
			port = 80;
#endif
		}

		if (iao_getembeddedvariable("PROXY_HOST", proxyhost)==0)
		{
			iao_logmsg("Using proxy settings");
			doproxy = 1;
			iao_getembeddedvariable("PROXY_PORT", buff);
			proxyport = atoi(buff);
			proxycridentials[0]=0;
			iao_getembeddedvariable("PROXY_CRIDENTIALS", proxycridentials);
		}

		first_api_call=0;
		return 0;

	}
	return 0;
}

int iao_init()
{
	return iao_firsttime();
}

static void iao_datetime(char *buff)
{
	time_t t = time(NULL);
	struct tm * ts = localtime(&t);
	sprintf(buff, "%04d-%02d-%02d %02d:%02d:%02d", 
		ts->tm_year+1900, ts->tm_mon+1, ts->tm_mday,
		ts->tm_hour, ts->tm_min, ts->tm_sec);
}

static void iao_buildpostbuff(iao_agentinfo *agent, iao_http_channel *http, char *buff, char *request)
{
	char td[128];
	iao_datetime(td);

	sprintf(buff, "<iaop pver=\"%d\" atype=\"%s\" aver=\"%s\" os=\"%s\" id=\"%s\" ip=\"%s\" host=\"%s\""
                      " time=\"%s\" uptime=\"%d\" iter=\"%d\">%s</iaop>", 
		IAOP_VERSION,
		agent->type,
		agent->version,
		OS,
		agent->id,
		http->connection.myip,
		http->connection.myhost,
		td,
		(int)(time(NULL)-iao_born_time),
		iao_iteration,
		request
		);
}

static int iao_peelnode(char *name, char *buff, int len)
{
	char o[256];
	char c[256];
	int clen;
	int olen;
	
	sprintf(o, "<%s>", name);
	sprintf(c, "</%s>", name);
	olen = (int) strlen(o);
	clen = (int) strlen(c);

	if (memcmp(buff, o, olen)!=0)
	{
		sprintf(c, "iao_peelnode: Can't find opening tag: %s", o);
		iao_logmsg(c);
		return -1;
	}
	if (memcmp(buff+len-clen, c, clen)!=0)
	{
		sprintf(o, "iao_peelnode: Can't find closing tag: %s", c);
		iao_logmsg(o);
		return -1;
	}
	len = len - olen - clen;
	memmove(buff, buff+olen, len);
	*(buff+len)=0;
	return len;
}

static void iao_error(char *p)
{
	char buff[1024];
	iao_peelnode("error", p, (int)strlen(p));
	sprintf(buff, "iao_error: server returned an error: %s", p);
	iao_logmsg(buff);
}

static char *mylastip=NULL;
static void remembermyip(iao_http_channel *http)
{
	static char ip1[20];
	static char ip2[20];

	char *ip = http->connection.myip;
	if (mylastip==NULL)
	{
		strcpy(ip1, ip);
		mylastip = ip1;
	}
	else
	{
		char *newplace = mylastip==ip1?ip2:ip1;
		strcpy(newplace, ip);
		mylastip = newplace;
	}
}
char *iao_getmylastip()
{
	return mylastip;
}

static int iao_post_blocked(iao_agentinfo *agent, char *request, char **buff)
{
	int len;
	char post_buff[10240];
	iao_http_channel http;
	char *die = "<directive>die:";
	static int consecutive_not_iaop_compliant = 0;
#ifdef IAO_SSL
	int secure = 1;
#else
	int secure = 0;
#endif

	if (iao_firsttime()!=0)
	{
		return -1;
	}

	if (doproxy==1)
	{
		if (iao_http_proxy_open(host, port, secure, proxyhost, 
                    proxyport, proxycridentials, &http)!=0)
		{
			return -1;
		}
	}
	else 
	{
		if (iao_http_open(host, port, secure, &http)!=0) 
		{
			return -1;
		}
	}

	remembermyip(&http);
	

	iao_buildpostbuff(agent, &http, post_buff, request);

	if (iao_http_post(&http, IAOP_SERVLET, post_buff, (int)strlen(post_buff))!=0)
	{
		iao_http_close(&http);
		return -1;
	}

	len = iao_http_getpage(&http, buff);
	iao_http_close(&http);

	if (len<0) 
	{
		return -1;
	}

        len = iao_peelnode("iaop", *buff, len);
        if (len==-1)
        {
                iao_logmsg(*buff);
                iao_logmsg("iao_post: response not IAOP compliant");
                free(*buff);
		if ((consecutive_not_iaop_compliant++)==2000)
		{	
			iao_logmsg("iao_post: response not IAOP compliant for 200 times in a row");
			iao_logmsg("iao_post: There could be a firewall issue.");
			iao_logmsg("iao_post: Bailing out.");
			exit(-1);
		}
                return -1;
        }
	consecutive_not_iaop_compliant = 0;

	if (strncmp(*buff, "<error>", 7)==0)
	{
		iao_error(*buff);
		free(*buff);
		return -1;
	}

	if (strncmp(*buff, die, strlen(die))==0)
	{
		iao_logmsg(*buff);
		iao_logmsg("iao_post: Was asked to die by server. Bailing out...");
		free(*buff);
		exit(-1);
	}
	return len;
}

static int iao_post(iao_agentinfo *agent, char *request, char **buff)
{
	int ret;

	iao_lockmutex(mx);
	ret = iao_post_blocked(agent, request, buff);
	iao_releasemutex(mx);
	return ret;
}

static void nextiteration()
{
	iao_lockmutex(mx);
	iao_iteration++;
	iao_releasemutex(mx);
}

int iao_register(iao_agentinfo *agent, char *result_agent_id, int skaa) 
{
	char *buff;
	int len;
	char *method = skaa?"<register skaa=\"true\"/>":"<register/>";

	len = iao_post(agent, method, &buff);
	if (len<0)
	{
		return -1;
	}

	if (strncmp("<id>", buff, 4)!=0)
	{
		iao_logmsg("iao_register: failed!");
		iao_logmsg(buff);
		free(buff);
		return -1;
	}

	strncpy(result_agent_id, buff+4, len-9);
	result_agent_id[len-9]=0;
	free(buff);
	nextiteration();
	return 0;
}

static int directive(iao_agentinfo *agent, char *inner, char *verb, char *x)
{
	char *buff;
    int len = iao_post(agent, inner, &buff);
	char *del;

    if (len<0)
    {
            return -1;
    }

    len = iao_peelnode("directive", buff, len);
    if (len<0)
    {
            free(buff);
            return -1;
    }

    del = strstr(buff, ":");
    if (del==NULL)
    {
            iao_logmsg("directive: malformed directive, no ':'");
            free(buff);
            return -1;
    }       
            
    *del=0;
    strcpy(verb, buff);
    if (strcmp(verb,"ERROR")==0)
    {
            iao_logmsg("directive: Got an ERROR from the server");
            iao_logmsg(del+1);
            return -1;
    }
    strcpy(x, del+1);
    free(buff);
	nextiteration();
    return 0;
}

int iao_keepalive(iao_agentinfo *agent, char *verb, char *x)
{
	return directive(agent, "<keepalive />", verb, x);
}

static char *addbuff(char *p, char *buff)
{
	strcat(p, buff);
	return p+strlen(buff);
}

static char *addalarm(char *p, iao_alarm *alarm)
{
	char *closetag;
	switch (alarm->type)
	{
	case 'C':
	case 'c':
		alarm->type = 'C';
		p = addbuff(p, "<clear ");
		closetag = "</clear>";
		break;
	case 'E':
	case 'e':
	case 'A':
	case 'a':
	default:
		if (alarm->type=='e'||alarm->type=='E')
		{
			p = addbuff(p, "<event ");
			closetag = "</event>";
		}
		else
		{
			p = addbuff(p, "<alarm ");
			closetag = "</alarm>";
		}
		switch(alarm->severity)
		{
		case 'M':
			p = addbuff(p, "severity=\"M\" ");
			break;
		case 'C':
			p = addbuff(p, "severity=\"C\" ");
			break;
		case 'm':
		default:
			p = addbuff(p, "severity=\"m\" ");
			break;
		}
		break;
	}

	p = addbuff(p, "resource=\"");
	p = addbuff(p, alarm->resource);
	p = addbuff(p, "\" id=\"");
	p = addbuff(p, alarm->id);
	p = addbuff(p, "\">");
	if (!(alarm->description[0]==0 && alarm->type=='C'))
	{
		p = addbuff(p, "<![CDATA[");
		p = addbuff(p, alarm->description);
		p = addbuff(p, "]]>");
	}
	p = addbuff(p, closetag);
	return p;
}

static void validatexml(char *buff)
{
	static char **transform=NULL;
	char *p = buff;
	if (transform==NULL)
	{
		transform = malloc(256*sizeof(char*));
		memset(transform, 0, 256*sizeof(char*));
		transform['&'] = "&amp;";
		transform['<'] = "&lt;";
		transform['\''] = "&apos;";
		transform['"'] = "&quot;";
		transform['>'] = "&gt;";
	}

	while (*p!=0)
	{
		if (transform[(int)(*p)]!=NULL)
		{
			char *tt = transform[(int)(*p)];
			int tlen = (int)strlen(tt);

			memmove(p+tlen, p+1, strlen(p));
			memcpy(p, tt, tlen);
			p+=(tlen-1);
		}
		p++;
	}
}

static void alarmslist(char *buff, iao_alarm alarms[], int count)
{
	char *p = buff+strlen(buff);
	int i;
	for (i=0;i<count;++i)
	{
		validatexml(alarms[i].id);
		validatexml(alarms[i].resource);
		p = addalarm(p, &alarms[i]);
	}
}

int iao_alarms(iao_agentinfo *agent, iao_alarm alarms[], int count,
               int inc, char *verb, char *x) 
{
	char alist[10240];

	iao_logmsg("iao_alarms: reporting agent state");
	sprintf(alist, "<alarms incremental=\"%s\">", inc?"true":"false");
	alarmslist(alist, alarms, count);
	strcat(alist, "</alarms>");

	return directive(agent, alist, verb, x);
}

int iao_alarms_ifchange(iao_agentinfo *agent, iao_alarm alarms[], int count,char *verb, char *x)
{
#define MAX_ALARM_BUFF_SIZE 50000
	static char *lasttime = NULL;
	static time_t lasttimetime = 0;
	time_t now = time(NULL);
	int effectivecount = 0;
	char thistime[MAX_ALARM_BUFF_SIZE]="";
	char *p = thistime;
	int i;

	for (i=0;i<count;++i)
	{
		if (alarms[i].type!='C')
		{
			p+=sprintf(p, "%s", alarms[i].id);
			p+=sprintf(p, "%s", alarms[i].resource);
			effectivecount++;
			if (p-thistime>MAX_ALARM_BUFF_SIZE-1000)
			{
				iao_logmsg("iao_alarms_ifchange: too many alarms");
				return iao_alarms(agent, alarms, count, 0, verb, x);
			}
		}
	}
	if (lasttime==NULL)
	{
		lasttime = (char *)malloc(MAX_ALARM_BUFF_SIZE);
		lasttime[0]=0;
	}
	if (strcmp(lasttime, thistime)==0 && (now-lasttimetime<60*60))
	{
		return 1;
	}
	else
	{
		int ret = iao_alarms(agent, alarms, count, 0, verb, x);
		if (ret==0)
		{
			strcpy(lasttime, thistime);
			lasttimetime = time(NULL);
		}
		return ret;
	}
}

static int buildreqvar(char *str, iao_confparam *param)
{
        return sprintf(str, "<param name=\"%s\" type=\"%s\" default=\"%s\"/>",
                param->name, param->type, param->defvalue);
}
static int buildreqvars(char *tag, char *str, iao_confparam *params)
{
        char *p = str;
        int i=0;

        p+=sprintf(p, "<%s>",tag);
        while (params+i!=NULL)
        {
		if ( (params+i)->name==NULL)
		{
			break;
		}
                p+=buildreqvar(p, params+i);
                i++;
        }
        p+=sprintf(p, "</%s>",tag);
        return (int) (p-str); 
}       

int iao_fetchconfig(iao_agentinfo *agent,
                    iao_confparam *agentconf, iao_confparam *resources)
{
        char buff[8192];
        char *p = buff;
	char *retbuff;
	int len;
        
        p+=sprintf(p, "<fetchconfig>");
        if (agentconf!=NULL)
        {
                p+=buildreqvars("agent", p, agentconf);
        }
        if (resources!=NULL)
        {
                p+=buildreqvars("resources", p, resources);
        }
        sprintf(p, "</fetchconfig>"); 

        len = iao_post(agent, buff, &retbuff);
        if (len<0)
        {
                return -1;
        }
        
        if (strncmp("<config>", retbuff, 8)!=0)
        {       
                iao_logmsg("iao_fetchconfig: failed!");
                iao_logmsg(retbuff);
                free(retbuff);
                return -1;
        }

		iao_lockmutex(mx);
		iao_parseconfig(retbuff, len);
		iao_releasemutex(mx);

        free(retbuff);
        return 0;

}




#define EVNC "!!!EMBEDDED_VARIABLE_NOT_CONFIGURED!!!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ENDOFBUFF"

static char *iao_getembededbuff()
{
	static char *embedded_variables = EVNC;
	return embedded_variables;
}

int iao_getembeddedvariable(char *name, char *result) 
{
	char *nstart;
	char *vstart;
	char *vend;
	size_t len;

	char nnn[IAOP_MAX_LEN];

	sprintf(nnn, "<NAME/>%s<VALUE/>", name);
	nstart = strstr(iao_getembededbuff(), nnn);
	if (nstart==NULL) 
	{
		return -1;
	}
	vstart = nstart+strlen(nnn);
	vend = strstr(vstart, "<END/>");
	if (vend==NULL)
	{
		return -1;
	}
	len = vend - vstart;
	strncpy(result, vstart, len);
	result[len] = 0;
	return 0;
}

static int advance(char *token, char **buff)
{
	int len = (int)strlen(token);
	if (strncmp(token, *buff, len)==0) 
	{
		(*buff) = *buff + len;
		return 0;
	}
	else
	{
		return -1;
	}
}

static void *mymemmem(void *big, int biglen, void *sub, int sublen)
{
	int i;
	for (i=0;i<biglen-sublen;i++)
	{
		if (memcmp( ((char*)big)+i, sub, sublen)==0)
		{
			return ((char*)big)+i;
		}
	}
	return NULL;
}

static void proxiate(char *buff, int len)
{
	char ph[IAOP_MAX_LEN];
	char pp[IAOP_MAX_LEN];

	if (iao_getembeddedvariable("PROXY_HOST", ph)==0)
	{
		char *srch = "<END/>@@@@@@@@@@@@@@@@@@@@@";
		int slen = (int)strlen(srch);
		char *found = mymemmem(buff, len, srch, slen);
		iao_getembeddedvariable("PROXY_PORT", pp);

		if (found!=NULL)
		{
			char addbuff[512];
			int addlen = sprintf(addbuff, "<NAME/>PROXY_HOST<VALUE/>%s<END/>"
			                              "<NAME/>PROXY_PORT<VALUE/>%s<END/>", ph, pp);
			found+=strlen("<END/>");
			memcpy(found, addbuff, addlen);
		}
		iao_getembeddedvariable("PROXY_PORT", pp);

		
		
	}
}


int iao_getbinary(iao_agentinfo *agent, void **buff, int *size)
{
	char authenticbuff[IAOP_MAX_LEN];
	char *authentic=NULL;
	int len;
	char *head1 = "<binary authentic=\"";
	char *head2 = "\">";
	char *tail = "</binary>";
	int tailen;
	char *p;
	void *bin;

	if (iao_getembeddedvariable("AUTHENTIC_TAG", authenticbuff)==0) 
	{
		authentic = authenticbuff;
	}
	else
	{
		iao_logmsg("iao_getbinary: AUTHENIC_TAG variable not embedded");
	}

	len = iao_post(agent, "<getbinary />", (char**)buff);
	if (len<0)
	{
		return -1;
	}

	p = *buff;

	if (advance(head1, &p)!=0)
	{
		iao_logmsg("iao_getbinary: <binary> tag invalid");
		free(*buff);
		return -1;
	}
	if (authentic)
	{
		if (advance(authentic, &p)!=0)
		{
			iao_logmsg("iao_getbinary: binary not authentic");
			free(*buff);
			exit(-1);
		}
	}
	else
	{
		p = strstr(p, head2);
	}

	if (advance(head2, &p)!=0)
	{
                iao_logmsg("iao_getbinary: <binary> tag  invalid (closing)");
                free(*buff);
                return -1;
	}

	tailen = (int) strlen(tail);
	if (strncmp(tail, ((char*)*buff)+len-tailen, tailen)!=0) 
	{
                iao_logmsg("iao_getbinary: </binary> tag invalid");
                free(*buff);
                return -1;
	}

	*(((char*)*buff)+len-tailen) = 0; /* terminate b64 with a null */

	bin = malloc(len);
	*size = iao_base64decode(bin, p);
	proxiate(bin, *size);
	
	free(*buff);
	*buff = bin;
	return 0;
}
 
static int isstandardparam(char *name) 
{
	return strcmp(name, IAO_AGENTPARAM_ALIAS)==0 ||
	       strcmp(name, IAO_AGENTPARAM_GROUP)==0 ||
	       strcmp(name, IAO_AGENTPARAM_NOTIF)==0 ||
	       strcmp(name, IAO_AGENTPARAM_SKAA)==0 ||
	       strcmp(name, IAO_AGENTPARAM_RESOURCES)==0;
}

int iao_setparams(iao_agentinfo *agent, iao_param namevalues[]) 
{
	char buffer[10000];
	char *p = buffer;
	iao_param *param = namevalues;
	int len;

	p+=sprintf(p, "<setparam>");

	while (param->name) 
	{
		char *isconfig = 
			isstandardparam(param->name) ?
			"":"config.";

		p+=sprintf(p, "<%s%s><![CDATA[%s]]></%s%s>",
			isconfig, param->name, param->value,
			isconfig, param->name);
		param++;
	}
	p+=sprintf(p, "</setparam>");

        len = iao_post(agent, buffer, &p);
        if (len<0)
        {
                return -1;
        }
	free(p);
	return 0;
}

