/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#pragma warning(disable : 4996)
#else
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#endif

#include "iao_comm.h"
#include "iao_threads.h"

#ifdef IAO_SSL
#include "cyassl/openssl/ssl.h"
#include "cyassl/ssl.h"
#include "cyassl/openssl/rand.h"
#include "cyassl/openssl/err.h"
static SSL_CTX *sslctx = NULL;
#endif

static iao_mutex ghbn_mutex = NULL;


static struct timeval gettimeval(int milisec)
{
   struct timeval t;
   t.tv_sec=milisec / 1000;
   t.tv_usec=(milisec % 1000) * 1000;
   return t;
}

static void iptostr(unsigned long ip, char *buff)
{	
	sprintf(buff, "%d.%d.%d.%d",
		((unsigned char *)(&ip))[0],
		((unsigned char *)(&ip))[1],
		((unsigned char *)(&ip))[2],
		((unsigned char *)(&ip))[3]
		);
}


static int readyto(int send, iao_connection *s, int milisec)
{
	/* Variables */
	struct timeval time_struct,*tm;
	int err;
	int fd = s->fd;

	/* Mask */
	fd_set tmp_Mask, tmp_ExceptMask;
	FD_ZERO(&tmp_Mask);
	FD_SET(fd,&tmp_Mask);
	FD_ZERO(&tmp_ExceptMask);
	FD_SET(fd,&tmp_ExceptMask);

	/* time */
	if (milisec==-1)
	{
		tm=NULL;
	}
	else
	{
		tm=&time_struct;
		time_struct=gettimeval(milisec);
	}

	/* and go! */
	if (send)
	{
		err = select( (unsigned int) fd+1,0,&tmp_Mask,&tmp_ExceptMask,tm);
	}
	else
	{
		err = select( (unsigned int) fd+1,&tmp_Mask,0,&tmp_ExceptMask,tm);
	}
	if (err<0)
	{
		iao_logsysmsg(send ? "readytosend: select returned an error" : "readytorecv: select returned an error");
		s->err = ERR_COMM_RESOURCE;
		return -1;
	}

	if (FD_ISSET(fd,&tmp_ExceptMask))
	{
		iao_logsysmsg(send ? "readytosend: disconnected [exception bit]" : "readytorecv: disconnected [exception bit]");
		s->err = ERR_COMM_DISCONNECTED;
		return -1;
	}
	if (!(FD_ISSET(fd,&tmp_Mask)))
	{
		iao_logmsg(send ? "readytosend: timeout" : "readytorecv: timeout");
		s->err = ERR_COMM_TIMEOUT;
		return -1;
	}
	return 0;
}

static int readytosend(iao_connection *s, int milisec)
{
	return readyto(1, s, milisec);
}

static int readytorecv(iao_connection *s, int milisec)
{
	return readyto(0, s, milisec);
}

long iao_gethostbyname(char *host, char *ip)
{
        struct hostent *hp;
	long ret = 0;
	static long ruonipcache = 0;
	
	iao_lockmutex(ghbn_mutex);
	hp=gethostbyname(host);
        if (hp!=NULL)
        {
		ret = *(long *)hp->h_addr;
	}
	if (strcmp(host, "agent.r-u-on.com")==0) /* if you want to monitor a DNS... */
	{
		if (ret==0)
		{
			ret = ruonipcache;
		}
		else
		{
			ruonipcache = ret;
		}
	}
	if (ip!=NULL)
	{
		iptostr(ret, ip);
	}
	iao_releasemutex(ghbn_mutex);
	return ret;
}

static void resetconnectstruct(iao_connection *connection)
{
	connection->fd = -1;
	connection->myhost[0]=0;
	connection->myip[0]=0;
	connection->remotehost[0]=0;
	connection->remoteip[0]=0;
	connection->err=ERR_COMM_OKAY;
    connection->myport = 0;
    connection->remoteport = 0;
#ifdef IAO_SSL
	connection->ssl = NULL;
#endif
	connection->udp = 0;
}
    
int iao_connect(char *host, int port, iao_connection *result)
{
	/* variables */
	struct in_addr i_addr;
	struct sockaddr_in client;
	struct sockaddr_in addr;
	int keepalive_flag=1;
	unsigned int addrlen = sizeof(struct sockaddr_in);

	iao_tcpinit(); /* has an effect only the first time */

	/* reset struct */
    resetconnectstruct(result);

	/* Host to ip */
	i_addr.s_addr = iao_gethostbyname(host, result->remoteip);
	if (i_addr.s_addr==0)
	{
		iao_logsysmsg("iao_connect: Can't translate host to ip");
		result->err=ERR_COMM_HOSTNAME;
		return -1;
	}

	strcpy(result->remotehost , host);


	/* make struct */
	client.sin_addr = i_addr;
	client.sin_family = AF_INET;                
	client.sin_port   = htons(port);           

	/* Get a socket */
	if ((result->fd = (int)socket(AF_INET,SOCK_STREAM,0)) == -1) 
	{
		iao_logsysmsg("iao_connect: Can't obtain file descriptor");
		result->err=ERR_COMM_RESOURCE;
		return -1;
	}

	/* Set keep alive */
	if (setsockopt(result->fd, SOL_SOCKET, SO_KEEPALIVE,
		(char *) &keepalive_flag, sizeof(keepalive_flag)) == -1)
	{
		iao_logsysmsg("iao_connect: Set keep alive on socket failed");
		result->err=ERR_COMM_RESOURCE;
		iao_close(result);
		return -1;
	}

	/* connect */
	if (connect( result->fd, (struct sockaddr *)&client, sizeof(client)) == -1)
	{
#ifdef WIN32
		int e = WSAGetLastError();
		WSASetLastError(e);
		result->err =(e==WSAECONNREFUSED?ERR_COMM_REFUSED:ERR_COMM_NO_CONNECTION);
#else
		result->err=(errno==ECONNREFUSED ? ERR_COMM_REFUSED : ERR_COMM_NO_CONNECTION);
#endif
		iao_logsysmsg("iao_connect: Remote host unreachable");
		iao_close(result);
		return -1;
	}

	if (getsockname(result->fd, (struct sockaddr *)&addr, &addrlen) != -1) 
	{
		iptostr(addr.sin_addr.s_addr, result->myip);
	}

	result->remoteport = port;

	if (gethostname(result->myhost,sizeof(result->myhost))!=0) 
	{
		strcpy(result->myhost, "*undetermined*");
	}
	return 0;
}

#ifdef IAO_SSL

static SSL_SESSION *agentsslsession = NULL;
static int doreuse(iao_connection *c) 
{
	return strcmp(c->remotehost, "agent.r-u-on.com")==0 || strcmp(c->remotehost, "ruondev")==0;
}

static int tryreuse(iao_connection *c) {
	if (doreuse(c) && agentsslsession!=NULL) 
	{
		SSL_set_session(c->ssl, agentsslsession);
		agentsslsession = NULL;
	}
	return 1;
}

static void saveforreuse(iao_connection *c)
{
	if (doreuse(c)) 
	{
		agentsslsession = SSL_get_session(c->ssl);
	}
}

int iao_sslify(iao_connection *result, int milisec)
{
	int stage = 0;
	char d[512];
	result->ssl = SSL_new(sslctx);
	if (result->ssl!=NULL)
	{
		CyaSSL_set_timeout(result->ssl, 24 * 60 * 60);
		stage = 1;
		if (SSL_set_fd(result->ssl, result->fd) && tryreuse(result))
		{
#ifdef WIN32
			int nonblock = 1;
			stage = 2;
			if (ioctlsocket(result->fd, FIONBIO, &nonblock)==0)
#else
			int flags = fcntl(result->fd, F_GETFL);
			int nonblock = flags|O_NONBLOCK;
			stage = 2;
			if (fcntl (result->fd, F_SETFL, nonblock)>=0)
#endif
			{
				int inprogress=1;
				stage = 3;
				while (inprogress)
				{
					int errret;
					int ret = SSL_connect(result->ssl);
					switch (ret)
					{
						case 1:
							saveforreuse(result);
							return 0;
						case 0:
							inprogress = 0;
							stage = 4;
							break;
						default:
							errret = SSL_get_error(result->ssl, ret);
							switch (errret)
							{
								case SSL_ERROR_WANT_READ:
									inprogress = readytorecv(result, milisec)==0;
									stage = 5;
									break;
								case SSL_ERROR_WANT_WRITE:
									inprogress = readytosend(result, milisec)==0;
									stage = 6;
									break;
								default:
									sprintf(d, "iao_sslconnect: SSL_get_error %d %d",
										 ret, errret);
									iao_logmsg(d);
									inprogress = 0;
									stage = 7;
									break;
									
							}
							break;
					}
				}
			}
		}
	}
	sprintf(d, "iao_sslconnect: SSL_connect failed (stage %d)", stage);
	iao_logsysmsg(d);
	result->err = ERR_SSL_FAILED;
	iao_close(result);
	return -1;
}

int iao_sslconnect(char *host, int port, iao_connection *result, int milisec)
{
	int ret = iao_connect(host, port, result);
	if (ret!=0)
	{
		return ret;
	}
	return iao_sslify(result, milisec);
}
#endif

static int udpsend(iao_connection *s, char *buff, int len)
{
	struct sockaddr_in sin;
	int err;
    	sin.sin_family = AF_INET;
    	sin.sin_port = htons(s->remoteport);
    	sin.sin_addr.s_addr = s->udp;
	err = sendto(s->fd, buff, len, 0, (struct sockaddr *) &sin, sizeof(sin));
	if (err==-1)
	{
		s->err = ERR_COMM_DISCONNECTED;
               	iao_logsysmsg("udpsend: sendto returned with an error");
	}
	return err;
}

int iao_send(iao_connection *s, char *buff, int len, int milisec)
{
	if (readytosend(s, milisec)==0) 
	{
		if (s->udp)
		{
			return udpsend(s, buff, len);
		}
#ifdef IAO_SSL
		if (s->ssl==NULL)
		{
#endif
			int err = send(s->fd,buff,len, 0 );
			if (err==-1)
			{
				s->err = ERR_COMM_DISCONNECTED;
				iao_logsysmsg("iao_send: send returned with an error");
				return -1;
			}
			else 
			{
				return err;
			}
#ifdef IAO_SSL
		}
		else
		{
			int err = SSL_write(s->ssl, buff, len);
			if (err<=0)
			{
				iao_logsysmsg("iao_send: SSL_write returned with an error");
				s->err = ERR_COMM_DISCONNECTED;
				return -1;
			}
			else
			{
				return err;
			}
		}
#endif
	}
	else
	{
		return -1;
	}
}

#ifdef IAO_SSL
static int sslrecv(iao_connection *s, char *buff, int len, int milisec)
{
	while (1)
	{
		/* first attempt (in buffer) */
		int err = SSL_read(s->ssl, buff, len); 
		if (err>0) 
		{
			return err;
		}
		else
		{
			if (SSL_get_error(s->ssl, err)==SSL_ERROR_WANT_READ)
			{
				err = readytorecv(s, milisec);
				if (err!=0)
				{
					return -1;
				}
			}
			else
			{
				s->err = ERR_COMM_DISCONNECTED;
				iao_logsysmsg("iao_recv: SSL_read returned with an error");
				return -1;
			}
		}
	}
}
#endif

int udprecv(iao_connection *s, char *buff, int len)
{
	struct sockaddr_in sin;
	unsigned int sinsize = sizeof(sin);
	int ret = recvfrom(s->fd, buff, len, 0, (struct sockaddr *)&sin, &sinsize);
	if (ret<=0)
	{
		s->err = ERR_COMM_DISCONNECTED;
	}
	return ret;
}

int iao_recv(iao_connection *s, char *buff, int len, int milisec)
{
#ifdef IAO_SSL
	if (s->ssl!=NULL)
	{
		return sslrecv(s, buff, len, milisec);
	}
#endif
	if (readytorecv(s, milisec)==0) 
	{
		if (s->udp)
		{
			return udprecv(s, buff, len);
		}
		else
		{
			int err = recv(s->fd,buff,len, 0 );
			if (err<=0)
			{
				s->err = ERR_COMM_DISCONNECTED;
				return -1;
			}
			else 
			{
				return err;
			}
		}
	}
	else
	{
		return -1;
	}
}

void iao_close(iao_connection *s)
{
#ifdef IAO_SSL
	if (s->ssl!=NULL)
	{
		if (SSL_shutdown(s->ssl)==0)
		{
			SSL_shutdown(s->ssl);
		}
		SSL_free(s->ssl);
	}
#endif
#if defined WIN32
	closesocket(s->fd);
#else
	close(s->fd);
#endif
}


int iao_tcpinit()
{
	if (ghbn_mutex==NULL) {
#if defined WIN32
	int ver_mjr=2;
	int ver_mnr=0;
	WORD wVersionRequested=MAKEWORD( ver_mjr, ver_mnr );
	WSADATA wsaData;
	int ret;
	if ((ret=WSAStartup( wVersionRequested, &wsaData ))!=0)
	{
		switch (ret)
		{
		case WSASYSNOTREADY:
			iao_logsysmsg("iao_tcpinit: Underlying network subsystem is not ready");
			break;
		case WSAVERNOTSUPPORTED:
			iao_logsysmsg("iao_tcpinit: WinSock version  not supported");
			break;
		case WSAEINPROGRESS:
			iao_logsysmsg("iao_tcpinit: A blocking Windows Sockets 1.1 operation is in progress");
			break;
		case WSAEPROCLIM:
			iao_logsysmsg("iao_tcpinit: Maximum WinSock tasks reached");
			break;
		case WSAEFAULT:
			iao_logsysmsg("iao_tcpinit: Invalid pointer");
			break;
		default:
			iao_logsysmsg("iao_tcpinit: Unknow reason");
			break;
		}
		return -1;
	}
#else
	/*sigignore(SIGPIPE); changed to a more portable call */
	signal(SIGPIPE, SIG_IGN);
#endif
#ifdef IAO_SSL
	SSL_library_init();
	sslctx = SSL_CTX_new(SSLv3_client_method());
	if (sslctx==NULL)
	{
		iao_logmsg("Failed creating ssl context");
		return -1;
	}
	SSL_CTX_set_verify(sslctx, SSL_VERIFY_NONE, NULL);

	/*
	if ever we want to clean up, call:
     	SSL_CTX_free(sslctx);
`	*/
#endif
	ghbn_mutex = iao_createmutex();
	if (ghbn_mutex==NULL)
	{
		iao_logmsg("Can't create gethostbyname mutex");
		return -1;
	}
	return 0;
} else
	{
		return 0;
	}
}

char *iao_comm_errstr(iao_connection *connection)
{
	static char *ss[] = {
        	"OK",
        	"Local server resource issue",
        	"Can't translate host name",
        	"No connection could be made to host",
        	"Connection actively refused by host",
        	"Timeout occured while negotiating with host",
        	"Host disconnected",
		"SSL negotiation failed"
	};
	return ss[-(connection->err)];
}



int iao_server(int port, iao_connection *result)
{
    struct sockaddr_in sin;
    unsigned int sinlen = sizeof(sin);
 
    resetconnectstruct(result);
    if ((result->fd = (int)socket(AF_INET, SOCK_STREAM,IPPROTO_TCP))<0) 
    {
        result->err = ERR_COMM_RESOURCE;
        iao_logsysmsg("iao_server: error creating socket");
        return -1;
    }

    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);
    sin.sin_addr.s_addr = INADDR_ANY;
    memset(&(sin.sin_zero), 0, sizeof(sin.sin_zero));

    if (bind(result->fd, (struct sockaddr*)&sin, sizeof(sin))!=0)
    {

        iao_logsysmsg("iao_server: bind failed");
        result->err = ERR_COMM_PORTINUSE;
        iao_close(result);
        return -1;
    }

    if (getsockname(result->fd, (struct sockaddr *)&sin, &sinlen)!=0)
    {
        iao_logsysmsg("iao_server: getsockname failed");
        result->err = ERR_COMM_RESOURCE;
        iao_close(result);
        return -1;
    }
    result->myport = ntohs(sin.sin_port);

    if (listen(result->fd, 64)!=0)
    {
        iao_logsysmsg("iao_server: listen failed");
        result->err = ERR_COMM_RESOURCE;
        iao_close(result);
        return -1;
    }
    return 0;
}

int iao_accept(iao_connection *server, iao_connection *client, int milisec)
{
	/* Variables */
	struct timeval time_struct,*tm;
	int err;
    struct sockaddr_in sin;
    unsigned int sockaddrlen = sizeof(sin);


    /* Mask */
	fd_set tmp_Mask, tmp_ExceptMask;
	FD_ZERO(&tmp_Mask);
	FD_SET(server->fd,&tmp_Mask);
	FD_ZERO(&tmp_ExceptMask);
	FD_SET(server->fd,&tmp_ExceptMask);

    resetconnectstruct(client);

    /* time */
	if (milisec==-1)
	{
		tm=NULL;
	}
	else
	{
		tm=&time_struct;
		time_struct=gettimeval(milisec);
	}

	err = select( (unsigned int) server->fd+1,&tmp_Mask,NULL,&tmp_ExceptMask,tm);
    	if (err<0)
	{
		iao_logsysmsg("iao_accept: select returned an error");
		server->err = ERR_COMM_RESOURCE;
		return -1;
	}

	if (FD_ISSET(server->fd,&tmp_ExceptMask))
	{
		iao_logsysmsg("iao_accept: socket closed");
		server->err = ERR_COMM_DISCONNECTED;
		return -1;
	}
	if (!(FD_ISSET(server->fd,&tmp_Mask)))
	{
		server->err = ERR_COMM_TIMEOUT;
		return -1;
	}

    /* Now accept */    
    client->fd = (int) accept(server->fd, (struct sockaddr*) &sin, &sockaddrlen);
    if (client->fd<0)
    {
        iao_logsysmsg("iao_accept: accept failed");
        return -1;
    }
    iptostr(sin.sin_addr.s_addr, client->remotehost);
	return 0;
}

int iao_udp(iao_connection *udp, int myport, char *host, int port) {

	struct sockaddr_in sin;
        unsigned int sinlen = sizeof(sin);
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(myport);
	sin.sin_addr.s_addr = INADDR_ANY;

	resetconnectstruct(udp);

	udp->udp = iao_gethostbyname(host, udp->remoteip);
	if (udp->udp==0)
	{
		iao_logsysmsg("iao_udp: Can't translate host to ip");
                udp->err=ERR_COMM_HOSTNAME;
		return -1;
	}

	if ((udp->fd = (int) socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
    	{
		iao_logsysmsg("iao_udp: Can't obtain file descriptor");
		udp->err = ERR_COMM_RESOURCE;
		return -1;
    	}
	
	if (bind(udp->fd, (struct sockaddr *) &sin, sizeof(sin))!=0)
	{
		iao_logsysmsg("iao_udp: bind failed");
		udp->err = ERR_COMM_PORTINUSE;
		iao_close(udp);
		return -1;
	}
	if (getsockname(udp->fd, (struct sockaddr *)&sin, &sinlen)!=0)
    	{
        	iao_logsysmsg("iao_udp: getsockname failed");
        	udp->err = ERR_COMM_RESOURCE;
        	iao_close(udp);
        	return -1;
    	}

    	udp->myport = ntohs(sin.sin_port);
	udp->remoteport = port;
	strcpy(udp->remotehost, host);
	return 0;
}



int bridgedata(iao_connection *rcv, fd_set *rcvmask, fd_set *exceptmask, iao_connection *snd, int millisec)
{
	if (FD_ISSET(rcv->fd, exceptmask))
	{
		return -1;
	}
	if (FD_ISSET(rcv->fd, rcvmask))
	{
		char buff[8192];
		int ret = iao_recv(rcv, buff, 8192, millisec);
		if (ret<0) 
		{
			return -1;
		}
		else if (ret>0)
		{
			iao_send(snd, buff, ret, millisec);
		}
	}
	return 0;
}

void iao_bridge(iao_connection *c1, iao_connection *c2, int millisec)
{
	fd_set rcvmask, exceptmask;

	FD_ZERO(&rcvmask);
	FD_ZERO(&exceptmask);

	while (1)
	{	
		int err;
		FD_SET(c1->fd, &rcvmask);
		FD_SET(c2->fd, &rcvmask);
		FD_SET(c1->fd, &exceptmask);
		FD_SET(c2->fd, &exceptmask);
		err = select((unsigned int) (c1->fd>c2->fd?c1->fd:c2->fd)+1,
			 &rcvmask, NULL, &exceptmask, NULL);
		if (err<0)
		{
			iao_logsysmsg("iao_bridge: select returned an error");
			return;
		}
		if (bridgedata(c1, &rcvmask, &exceptmask, c2, millisec) ||
		    bridgedata(c2, &rcvmask, &exceptmask, c1, millisec))
		{
			iao_logmsg("iao_bridge: session disconnected");
			return;
		}
	}
}
