/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#pragma warning(disable : 4996)
#endif 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "iao_http.h"
#include "iao_comm.h"

#define TIMEOUT  60000

static int iao_get_header(iao_http_channel* channel, char *buff);


static void anerror(iao_http_channel *channel, char *s)
{
	channel->statuscode = -1;
	strcpy(channel->status, s);
}

static void commerror(iao_http_channel *channel)
{
	channel->statuscode = channel->connection.err;
	strcpy(channel->status, iao_comm_errstr(&(channel->connection)));
}

int iao_http_open(char *host, int port, int secure, iao_http_channel *channel)
{
	int ret;
	channel->statuscode = 0;
	channel->status[0]=0;
	channel->proxycridentials[0]=0;
	channel->realhost[0]=0;
	channel->realport = 0;
	channel->contenttype[0] = 0;
	if (secure)
	{
#ifdef IAO_SSL
		ret = iao_sslconnect(host, port, &(channel->connection), TIMEOUT);
#else
		iao_logmsg("iao_http_connect: Can't provide secure connection - not linked with SSL");
		anerror(channel, "SSL not supported by agent");
		return -1;
#endif
	}
	else
	{
		ret =  iao_connect(host, port, &(channel->connection));
	}
	if (ret!=0)
	{
		commerror(channel);
        }

	return ret;
}


#ifdef IAO_SSL
static int proxyconnect(iao_http_channel *channel, char *host, int port, char *proxycridentials)
{
	int ret;
	char buff[10000];
	char *p = buff;
	int len;
	
	p+=sprintf(p, "CONNECT %s:%d HTTP/1.0\r\n", host, port);
        p+=sprintf(p, "User-Agent: ruon/1.0\r\n");


	if (proxycridentials[0]!=0)
	{
		p+=sprintf(p, "Proxy-Authorization: Basic %s\r\n", proxycridentials);
	}
	p+=sprintf(p, "\r\n");
	len = (int)(p - buff);
	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret==len)
	{
		if (iao_get_header(channel, buff)==0)
		{
			if (channel->statuscode==200)
			{
				ret = iao_sslify(&(channel->connection), TIMEOUT);
				if (ret==0)
				{
					return ret;
				}
				commerror(channel);
			}
			else
			{
				sprintf(buff, "proxyconnect: bad http status %d '%s'", 
					channel->statuscode, channel->status);
				iao_logmsg(buff);
				anerror(channel, buff);
			}
		}
		else
		{
			commerror(channel);
			iao_logmsg("proxyconnect: Error receiving header");
		}
	}
	else
	{
		commerror(channel);
		iao_logmsg("proxyconnect: error sending CONNECT request");
	}
	iao_close(&(channel->connection));
	return -1;
}
#endif

int iao_http_proxy_open(char *host, int port, int secure, char *proxyhost, int proxyport, char *proxycridentials, iao_http_channel *channel)
{
	int ret = iao_http_open(proxyhost, proxyport, 0, channel);
	if (ret==0)
	{
		if (secure)
		{
#ifdef IAO_SSL
			return proxyconnect(channel, host, port, proxycridentials);
#else
			iao_logmsg("iao_http_proxy_open: Can't provide secure connection - not linked with SSL");
			anerror(channel, "Can't provide a secure connection - not linked with SSL");
			return -1;
#endif
		}
		else
		{
			if (proxycridentials[0]!=0) 
			{
				strcpy(channel->proxycridentials, proxycridentials);
			}
			strcpy(channel->realhost, host);
			channel->realport = port;
			return 0;
		}
	}
	else
	{
		iao_logmsg("iao_http_proxy_open: Can't connect to proxy");
		commerror(channel);
		return ret;
	}
}


static void httpheader(char *buff, iao_http_channel* channel, char *url, int contentlen)
{
	char *method = contentlen>0?"POST":"GET";

	/* first line */
	if (channel->realhost[0]==0)
        { 
		buff+=sprintf(buff, "%s %s HTTP/1.0\r\n", method, url);
        }
	else
	{
		buff+=sprintf(buff, "%s http://%s:%d%s HTTP/1.0\r\n", method, channel->realhost, channel->realport, url);
	}

	/* User Agent */
	buff+=sprintf(buff, "User-Agent: ruon/1.0\r\n");

	/* Host */
	buff+=sprintf(buff, "Host: %s", channel->connection.remotehost);
	if (channel->connection.remoteport!=80 && channel->connection.remoteport!=443)
	{
		buff+=sprintf(buff, ":%d", channel->connection.remoteport);
	}
	buff+=sprintf(buff, "\r\n");

	/* Keepalive */
	buff+=sprintf(buff, "Connection: Keep-Alive\r\n");

	/* Content Length */
	if (contentlen>0)
	{
		buff+=sprintf(buff, "Content-Length: %d\r\n", contentlen);
	}

	/* Proxy basic authentication */
	if (channel->proxycridentials[0]!=0)
	{
		buff+=sprintf(buff, "Proxy-Authorization: Basic %s\r\n", channel->proxycridentials);
	}

	sprintf(buff, "\r\n");
}

int iao_http_get(iao_http_channel* channel, char *url)
{
	char buff[1024];
	int len;
	int ret;

	httpheader(buff, channel, url, -1);
	len = (int)strlen(buff);

	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret==len) 
	{
		return 0;
	}
	else
	{
		commerror(channel);
		return -1;
	}
}

int iao_http_post(iao_http_channel* channel, char *url, char *data, int datalen)
{
	char buff[1024];
	int len;
	int ret;

	httpheader(buff, channel, url, datalen);
	len = (int)strlen(buff);
	ret = iao_send(&(channel->connection), buff, len, TIMEOUT);
	if (ret!=len) 
	{
		commerror(channel);
		return -1;
	}
	ret=iao_send(&(channel->connection), data, datalen, TIMEOUT);
	if (ret==datalen)
	{
		return 0;
	}
	else
	{
		commerror(channel);
		return -1;
	}
}

static int iao_getbetween(char *start, char *end, char *buff, char *result)
{
	char *es;
	int startlen;
	int resultlen;
	char *ss = strstr(buff, start);
	if (ss==NULL)
	{
		return -1;
	}
	es = strstr(ss, end);
	if (es==NULL) 
	{
		return -1;
	}
	startlen  = (int) strlen(start);
	resultlen = (int) (es-ss-startlen);
	strncpy(result, ss+startlen, resultlen);
	result[resultlen] = 0;
	return 0;
}


static int iao_getattribute(char *attribname, char *buff, char *result)
{
	char b[256];
	strcpy(b, attribname);
	strcat(b, ": ");
	return iao_getbetween(b, "\r\n", buff, result);
}

static int iao_get_header(iao_http_channel* channel, char *buff)
{
	char *eoh = "\r\n\r\n";
	int maxlen=9999;
	int received = 0;
	int eohlen=(int)strlen(eoh);
	int eohrcv=0;
	char b;
	buff[received]=0;
	while (1)
	{
		int ret;
		if (received==maxlen)
		{
			anerror(channel, "Header too big");
			iao_logmsg("iao_get_header: Header too big");
			return -1;
		}
		ret = iao_recv(&(channel->connection), buff+received, 1, TIMEOUT);
		if (ret<=0)
		{
			fflush(stdout);
			iao_logmsg("get_header: disconnected while reading header");
			commerror(channel);
			return -1;
		}
		b = buff[received];
		/*if (b<' ') printf("[%d]", b ); else printf("%c", b);*/
		received++;
		if (received==4 && strncmp("HTTP", buff, 4)!=0) /* not HTTP header */
		{
		 	channel->statuscode = 200;
                        strcpy(channel->status, "OK");
			buff[4] = 0;
			return 0;
		}
		eohrcv = b==eoh[eohrcv] ? eohrcv+1 : 0; /* reset or advance */
		if (eohrcv==eohlen)
		{
			char *sp1, *sp2, *nl;
			char tmp[1024];
			int len;
			char *err = NULL;

			buff[received]=0;
			if ((sp1=strstr(buff, " "))==NULL) { err = "iao_get_header: Can't find space"; }
			if ((sp2=strstr(sp1+1, " "))==NULL) { err = "iao_get_header: Can't find second space"; }
			if ((nl=strstr(buff, "\r\n"))==NULL) { err = "iao_get_header: Can't find newline"; }
			if (err!=NULL)
			{
				iao_logmsg(err);
				anerror(channel ,"Error parsing header");
				return -1;
			}
			len = (int)(sp2-sp1-1);
			strncpy(tmp, sp1+1, len);
			tmp[len]=0;
			channel->statuscode = atoi(tmp);
			len = (int)(nl-sp2-1);
			strncpy(channel->status, sp2+1, len);
			channel->status[len]=0;
			return 0;
		}
	}
}

static int recv2disconnect(iao_http_channel* channel, char **result, char *header)
{
	int alloc = 0;
	int rcvd = 0;
	char b[1024];
	char *buff=NULL;
	int ret;

	if (strncmp(header, "HTTP", 4)!=0) // this is not header but data
	{
		buff = (char*)malloc(10000);	
		alloc = 10000;
		strcpy(buff, header);
		rcvd = (int)strlen(buff);
	}

	do {
		if ((ret = iao_recv(&(channel->connection), b, 1024, TIMEOUT))>0)
		{
			if (ret+rcvd>alloc)
			{
				char *t = (char *) malloc(alloc+10000);
				if (alloc>0)
				{
					memcpy(t, buff, alloc);
					free(buff);
				}
				buff = t;
				alloc+=10000;
			}
			memcpy(buff+rcvd, b, ret);
			rcvd+=ret;
		}
	} while (ret>0);

	if (channel->connection.err!=ERR_COMM_TIMEOUT)
	{
		if (rcvd==0)
                {
                       buff = (char *)malloc(1);
                }

		buff[rcvd]=0;
		*result = buff;
		return rcvd;
	}
	else
	{
		commerror(channel);
		free(buff);
		return -1;
	}
	
}


int iao_http_getpage(iao_http_channel* channel, char **buff)
{
	int received=0;
	int page_size;
	char cl[50];
	char header[10000];

	memset(header, 0, sizeof(header));	

	if (iao_get_header(channel, header)!=0)
	{
		return -1;
	}

	iao_getattribute("Content-Type", header, channel->contenttype);
	if (iao_getattribute("Content-Length", header, cl)!=0)
	{
		return recv2disconnect(channel, buff, header);
	}
	page_size = atoi(cl);

	if (page_size<0)
	{
		anerror(channel, "Negative page size");
		return -1;
	}

	(*buff) = malloc(page_size+1);
	(*buff)[0]=0;
	if (page_size==0)
	{
		return 0;
	}
	while (1)
	{
		int ret = iao_recv(&(channel->connection), (*buff)+received, page_size-received, TIMEOUT);
		if (ret<=-1)
		{
			free(*buff);
			commerror(channel);
			return -1;
		}
		if (ret==0)
		{
			if (received==page_size)
			{
				return received;
			}
			else
			{
				char eb[256];
				free(*buff);
                               sprintf(eb, "Partial data received (%d/%d bytes)", received, page_size); 
				anerror(channel, eb);
				return -1;
			}
		}

		received+=ret;
		(*buff)[received]=0;
		if (received>=page_size)
		{
			return received;
		}
	}
	free(*buff);
	commerror(channel);
	return -1;
}


void iao_http_close(iao_http_channel* channel)
{
	iao_close(&(channel->connection));
}

char *iao_http_status(int code)
{
	typedef struct {
		int code;
		char *s;
	} entry;
	
	static entry codes[] = {
                {100, "Continue"}, {101, "Switching Protocols"}, {200, "OK"},
                {201, "Created"}, {202, "Accepted"}, {203, "Non-Authoritative Information"},
                {204, "No Content"}, {205, "Reset Content"}, {206, "Partial Content"},
                {300, "Multiple Choices"}, {301, "Moved Permanently"}, {302, "Moved Temporarily"},
                {303, "See Other"}, {304, "Not Modified"}, {305, "Use Proxy"},
                {306, "Unused"}, {307, "Temporary Redirect"}, {400, "Bad Request"},
                {401, "Unauthorized"}, {402, "Payment Required"}, {403, "Forbidden"},
                {404, "Not Found"}, {405, "Method Not Allowed"}, {406, "Not Acceptable"},
                {407, "Proxy Authentication Required"}, {408, "Request Time-out"}, {409, "Conflict"},
                {410, "Gone"}, {411, "Length Required"}, {412, "Precondition Failed"},
                {413, "Request Entity Too Large"}, {414, "Request-URI Too Large"}, {415, "Unsupported Media Type"},
                {416, "Requested Range Not Satisfiable"}, {417, "Expectation Failed"}, {500, "Internal Server Error"},
                {501, "Not Implemented"}, {502, "Bad Gateway"}, {503, "Service Unavailable"},
                {504, "Gateway Time-out"}, {505, "HTTP Version not supported"},
		{0,0}
	};

	entry *p = codes;
	while (p->code!=0)
	{
		if (p->code==code)
		{
			return p->s;
		}
		p++;
	}
	
	switch (code/100)
	{
		case 1:
			return "1xx Intermediate status";
		case 2:
			return "2xx Successful response";
		case 3:
			return "3xx Redirect";
		case 4:
			return "4xx Request error";
		case 5:
			return "5xx Server error";
		default:
			return "Unknown HTTP status code";
	}
}

