/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include "ciao.h"
#include "cfw.h"
#include "plist.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
#include <windows.h>
#pragma warning(disable:4996)
#endif

#define PROC_OKAY	0
#define PROC_ABOVEMAX 	1
#define PROC_BELOWMIN 	2
#define PROC_MONITOR	3


int agentinit(iao_agentinfo *agent)
{
	return 15;
}

int agentmain(iao_agentinfo *agent, char *verb, char *x, int mode);

agentmetadata getagentinfo()
{
        static agentmetadata meta;

	static iao_confparam resourceparams[] = {
        	{ "process",  "String", "exe" },
        	{ "min",  "Integer", "1" },
        	{ "max",  "Integer", "1" },
		{0}
	};

        meta.Name = "ProcessOn";
        meta.name = "processon";
        /* When writing your own private agent don't add the ruon prefix to its type name. */
        /* This way you are certain to maintain name uniqueness. */
        meta.type = "ruon.ProcessOn";
#ifdef WIN32
        meta.version = "1.5";
#else
#ifdef Darwin
        meta.version = "1.5";
#else
        meta.version = "1.4";
#endif
#endif
        meta.agentparams = NULL;
        meta.resourceparams = resourceparams;
        meta.agentregister = NULL;
        meta.agentinit = agentinit;
        meta.agentmain = agentmain;

        return meta;
}

#define MAX_PROCESSES 256

typedef struct processnum_struct {
        char process[256];
	int max;
	int min;
        int reported;
        int actual;
} processnum;


static processnum processes[MAX_PROCESSES];
static int size = 0;

static void checkconfig()
{
	/* Reset configuration */
	char *p = iao_configresourcelist();
	size = 0;
	if (p!=NULL)
	{
		while (*p!=0 && size<MAX_PROCESSES)
		{
			int len = sprintf(processes[size].process, "%s", p);
			processes[size].reported = -1;
			processes[size].max = atoi(iao_configresourcevalue(p, "max"));
			processes[size].min = atoi(iao_configresourcevalue(p, "min"));
			size++;
			p+=(len+1);
		}
	}
}

static void resetcount()
{
	int i=0;
	while (i<size)
	{
		processes[i++].actual = 0;
	}
}

static int processcount()
{
	int i;
	char **l = proclist();
	if (l==NULL)
	{
		return -1;
	}
	resetcount();
	while (*l!=NULL)
	{
		i=0;
		for (i=0;i<size;++i)
		{
			if (match(processes[i].process, *l))
			{
				processes[i].actual++;
			}
		}
		l++;
	}
	return 0;
}

int calculatealarmid(int i)
{
        if (processes[i].actual>processes[i].max && processes[i].max>0)
        {
                return PROC_ABOVEMAX;
        }
        else if (processes[i].actual<processes[i].min)
        {
                return PROC_BELOWMIN;
        }
        else if (processes[i].actual>0 && processes[i].min==0 && processes[i].max==0)
        {
                return PROC_MONITOR;
        }
	 else 
	{
		return PROC_OKAY;
	}
}

char *alarmidstring(int alarmid)
{
	switch (alarmid)
	{
	case PROC_ABOVEMAX:
		return "PROC_ABOVEMAX";
	case PROC_BELOWMIN:
		return "PROC_BELOWMIN";
	case PROC_MONITOR:
		return "PROC_MONITOR";
	default:
		return "PROC_NOSUCHALARM";
	}
}

void alarmdesc(int alarmid, int i, char *desc)
{
	switch (alarmid) 
	{
	case PROC_ABOVEMAX:
		sprintf(desc, "Too many process instances (%d/%d)", processes[i].actual, processes[i].max);
		break;
	case PROC_BELOWMIN:
		if (processes[i].actual==0)
		{
			sprintf(desc, "Process is down");
		}
		else
		{
			sprintf(desc, "Too few process instances (%d/%d)", processes[i].actual, processes[i].min);
		}
		break;
	case PROC_MONITOR:
		if (processes[i].actual==1)
		{
			sprintf(desc, "Process is up");
		}
		else
		{
			sprintf(desc, "Process is up (%d instances)", processes[i].actual);
		}
		break;
	default:
		sprintf(desc, "Okay");
		break;
	}
}

static char alarmsev(int alarmid)
{
	switch (alarmid)
	{
		case PROC_ABOVEMAX:
			return 'M';
		case PROC_BELOWMIN:
			return 'C';
		case PROC_MONITOR:
		default:
			return 'm';
	}
}

static int report(iao_agentinfo *agent, char *verb, char *x)
{
	iao_alarm alarms[MAX_PROCESSES];
	int ac = 0;
	int i;
	int ret;

	for (i=0;i<size;++i)
	{
		int alarmid = calculatealarmid(i);
		if (alarmid!=PROC_OKAY)
		{
			strcpy(alarms[ac].resource, processes[i].process);
			strcpy(alarms[ac].id, alarmidstring(alarmid));
			alarmdesc(alarmid, i, alarms[ac].description);
			alarms[ac].type = 'A';
			alarms[ac].severity = alarmsev(alarmid);
			ac++;
		}
	}
	ret = iao_alarms(agent, alarms, ac, 0, verb, x);
	if (ret==0)
	{
		for (i=0;i<size;i++)
		{
			processes[i].reported = processes[i].actual;
		}
	}
	return ret;
}

static int same(int a, int b, int min, int max)
{
	/* if within range, no need to report changes */
	return (a>=min && a<=max && b>=min && b<=max) || a==b;
}

int agentmain(iao_agentinfo *agent, char *verb, char *x, int mode)
{
	int configchanged = mode&MODE_CONFIG_CHANGED;
	int statechanged = 0;
	if (configchanged)
	{
		checkconfig();
	}

	if (processcount()==-1)
	{
		exit(-1);
	}
	
	if (!configchanged)
	{
		int i=0;
		while (i<size)
		{
			if (!same(processes[i].actual, processes[i].reported, 
			          processes[i].min,    processes[i].max))
			{
				statechanged = 1;
				break;
			}
		i++;
		}
	}

	if (statechanged || configchanged)
	{
		return report(agent, verb, x);
	}
	else
	{
		return 1;
	}
	return 1;
}
